#!/usr/bin/python3

import pop.hub


def pop_seed():
    CONFIG = {
            'seed_name': {
                'positional': True,
                'help': 'The name of the project that is being created',
                },
            }

    hub = pop.hub.Hub()
    hub.pop.sub.add('pop.mods.conf')
    hub.opts = hub.conf.reader.read(CONFIG)
    hub.pop.seed.new()
